{% extends "!layout.html" %}
{% set html5_doctype = True %}
{% set script_files = script_files + ['_static/instaloaderdoc.js'] %}
{% block htmltitle %}
  {% if pagename == "index" %}
    <title>{{ title|striptags|e }}</title>
  {% else %}
    <title>{{ title|striptags|e }}{{ titlesuffix }}</title>
  {% endif %}
{% endblock %}
{% block extrahead %}
  {{ super() }}
  <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
  <meta name="google-site-verification" content="FNUsDdkOJM9obm9QEdrNBVd9AtAzHzA-aaSMCUrS6C8"/>
  {% if pagename == "index" %}
    <link rel="canonical" href="https://instaloader.github.io/"/>
  {% else %}
    <link rel="canonical" href="https://instaloader.github.io/{{ pagename }}.html"/>
  {% endif %}
{% endblock %}
{% block header %}
  <nav class="navbar navbar-light navbar-expand bg-light border-bottom">
    <a class="navbar-brand" href="{{ pathto(master_doc) }}">
      <img src="{{ pathto('_static/logo.png', 1) }}" width="30" height="30" class="d-inline-block align-top" alt="logo">
      <span class="ml-3">Instaloader</span>
    </a>
    <ul class="navbar-nav mr-auto">
      <li class="nav-item ml-0 ml-sm-2">
        <a class="nav-link"
           href="https://github.com/instaloader/instaloader/releases/tag/v{{ current_release }}"
           title="Released on {{ current_release_date }}" data-toggle="tooltip">
          v<b>{{ current_release }}</b>
        </a>
      </li>
    </ul>
    <ul class="navbar-nav ml-auto d-none d-sm-flex">
      <li class="nav-item">
        <a class="nav-link" href="https://github.com/instaloader/instaloader/issues">
          Issues
        </a>
      </li>
      <li class="nav-item">
        <a class="nav-link" href="https://github.com/instaloader/instaloader/releases">
          Releases
        </a>
      </li>
      <li class="nav-item">
        <a class="nav-link" href="https://github.com/instaloader/instaloader">
          GitHub
        </a>
      </li>
    </ul>
    <button class="d-inline-block d-md-none btn btn-outline-primary btn-sm" type="button"
            data-toggle="collapse" data-target="#instaloaderdocSidebar" aria-controls="instaloaderdocSidebar"
            aria-expanded="false" aria-label="Toggle navigation">
      <span class="navbar-toggler-icon"></span>
    </button>
  </nav>
{% endblock %}
{% block relbar1 %}{% endblock %}
{% block relbar2 %}{% endblock %}
{% block footer %}{% endblock %}
{% block content %}
  <div class="container-fluid">
    <div class="row">
      <div class="instaloaderdoc-sidebar collapse d-md-block col-md-3 col-xl-2 bg-light"
           id="instaloaderdocSidebar">
        <nav role="navigation" aria-label="Table of Contents">
          <form action="{{ pathto('search') }}" method="get" role="search" class="border-bottom mr-2 pb-2 mb-2 px-1">
            <label class="sr-only" for="searchQueryInput">Search</label>
            <input type="search" name="q" id="searchQueryInput" class="form-control my-2"
                   placeholder="Search..." aria-label="Search">
          </form>
          <h4>Contents</h4>
          {{ toctree(maxdepth=-1) }}
        </nav>
      </div>
      <main class="col-12 col-md-9 col-xl-10 pt-2" role="main">
        {% block body %}{% endblock %}
      </main>
    </div>
  </div>
{% endblock %}
