{% extends "!search.html" %}

{% block body %}
  <div id="fallback" class="admonition warning">
    <script type="text/javascript">$('#fallback').hide();</script>
    <p>
      {% trans %}Please activate JavaScript to enable the search
        functionality.{% endtrans %}
    </p>
  </div>
  {% if search_performed %}
    <h1>{{ _('Search Results') }}</h1>
    {% if not search_results %}
      <p>{{ _('Your search did not match any documents. Please make sure that all words are spelled correctly and that you\'ve selected enough categories.') }}</p>
    {% endif %}
  {% endif %}
  <div id="search-results">
    {% if search_results %}
      <ul>
        {% for href, caption, context in search_results %}
          <li><a href="{{ pathto(item.href) }}">{{ caption }}</a>
            <div class="context">{{ context|e }}</div>
          </li>
        {% endfor %}
      </ul>
    {% endif %}
  </div>
{% endblock %}